/*
** Module   : SavepointReversible.java
** Abstract : Common reversible based on java.sql.Savepoint interface
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 VMN 20130723          Created initial version
*/

package com.goldencode.p2j.persist;

import org.hibernate.Session;
import org.hibernate.jdbc.Work;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Savepoint;

/**
 * Common reversible based on java.sql.Savepoint interface.
 * <p>
 * <b>Implementation Notes</b><br>
 * Implementation based on two methods: setSavepoint(...) and rollbackSavepoint(). Both methods
 * use Hibernate Work interface to set/rollback Savepoint to Hibernate connection.
 *
 * Savepoint could give some problems for the Hibernate second-level cache or for dealing with
 * any records currently held in buffers. Also Savepoint reversible approach should be tested
 * for each database type.
 * <p>
 */
public class SavepointReversible extends Reversible
{
   /** DMO interface short name. */
   private String dmoName;

   /** Operation name corresponding to this reversible. It would be set as savepoint name. */
   private String operationName;

   /** Savepoint instance. */
   private Savepoint savepoint;

   /** The session which is invoking reversible operation. */
   private Session session;

   /**
    * Constructor.
    *
    * @param   dmoName
    *          DMO interface short name.
    * @param   operationName
    *          Operation name corresponding to this reversible. It would be set as savepoint
    *          name.
    * @param   session
    *          The session which is invoking reversible operation.
    */
   public SavepointReversible(String dmoName,
                              String operationName,
                              Session session)
   throws PersistenceException
   {
      this.operationName = operationName;
      this.dmoName = dmoName;
      this.session = session;
      setSavepoint();
   }

   /**
    * Get a string representation of the internal state of this object,
    * primarily for debugging purposes.
    *
    * @return  String representing this object's state.
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder(operationName);
      buf.append(" of ");
      buf.append(dmoName);

      return buf.toString();
   }

   /**
    * Indicate whether this <code>Reversible</code> implementation stores its
    * DMO locally, as opposed to storing an identifier and having to retrieve
    * the DMO from the backing database, or not making a DMO available at all.
    *
    * @return  <code>true</code> if the DMO is stored locally, else
    *          <code>false</code>.
    */
   protected boolean storesDMO()
   {
      return false;
   }

   /**
    * Get the unqualified interface name of the DMO being managed by this
    * reversible.
    *
    * @return  DMO interface name.
    */
   protected String getDMOName()
   {
      return dmoName;
   }

   /**
    * No-op.
    *
    * @return  Nothing.
    *
    * @throws  UnsupportedOperationException
    *          No DMO is required.
    */
   protected Serializable getId()
   {
      throw new UnsupportedOperationException(
              operationName + " does not require a DMO for persistable " + getDMOName());
   }

   /**
    * No-op.
    *
    * @return  Nothing.
    *
    * @throws  UnsupportedOperationException
    *          always;  no DMO is required.
    */
   protected Persistable getDMO()
   {
      throw new UnsupportedOperationException(
              operationName + " does not require a DMO for persistable " + getDMOName());
   }

   /**
    * Set the underlying DMO for this reversible action. This implementation is a no-op.
    *
    * @param   dmo
    *          The underlying DMO.
    */
   protected void setDMO(Persistable dmo)
   {
   }

   /**
    * Reverse operation.
    *
    * @param   rollForward
    *          If <code>true</code>, this method is being invoked in order
    *          to re-roll forward a non-undoable change, after a
    *          database-level rollback.  This indicates the modification is
    *          made within a new transaction invoked after the
    *          database-level rollback.
    *          If <code>false</code>, this method is being invoked in order
    *          to roll back an undoable change within the context of an
    *          application-level, full transaction or sub-transaction
    *          rollback.
    * @param   transaction
    *          <code>true</code> if this rollback occurs at a full
    *          transaction boundary, else <code>false</code>.
    *
    * @throws  PersistenceException
    *          if an error occurs at the database or in the persistence
    *          layer when updating the records.
    */
   protected void rollback(boolean rollForward, boolean transaction)
           throws PersistenceException
   {
      rollbackSavepoint();
   }

   /**
    * Create savepoint for the connection managed by this session.
    */
   private void setSavepoint()
   {
      session.doWork(new Work()
      {
         public void execute(Connection connection) throws SQLException
         {
            savepoint = connection.setSavepoint(operationName);
         }
      });
   }

   /**
    * Rollback savepoint for the connection managed by this session.
    */
   private void rollbackSavepoint()
   {
      session.doWork(new Work()
      {
         public void execute(Connection connection) throws SQLException
         {
            connection.rollback(savepoint);
         }
      });
   }
}
